# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import copy
import importlib
import logging
import os
import re
import sys
import time
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
import scipy.sparse
import sklearn.pipeline
from sklearn.metrics import accuracy_score, check_scoring, log_loss
from sklearn.model_selection import StratifiedKFold
from sklearn.utils.metaestimators import _safe_split

import lale.datasets.data_schemas

try:
    import torch

    torch_installed = True
except ImportError:
    torch_installed = False

from importlib import util

spark_loader = util.find_spec("pyspark")
spark_installed = spark_loader is not None
if spark_installed:
    from pyspark.sql.dataframe import DataFrame as spark_df

logger = logging.getLogger(__name__)

LALE_NESTED_SPACE_KEY = "__lale_nested_space"


def make_nested_hyperopt_space(sub_space):
    return {LALE_NESTED_SPACE_KEY: sub_space}


def assignee_name(level=1) -> Optional[str]:
    tb = traceback.extract_stack()
    file_name, line_number, function_name, text = tb[-(level + 2)]
    try:
        tree = ast.parse(text, file_name)
    except SyntaxError:
        return None
    assert tree is not None and isinstance(tree, ast.Module)
    if len(tree.body) == 1:
        stmt = tree.body[0]
        if isinstance(stmt, ast.Assign):
            lhs = stmt.targets
            if len(lhs) == 1:
                res = lhs[0]
                if isinstance(res, ast.Name):
                    return res.id
    return None


def arg_name(pos=0, level=1) -> Optional[str]:
    tb = traceback.extract_stack()
    file_name, line_number, function_name, text = tb[-(level + 2)]
    try:
        tree = ast.parse(text, file_name)
    except SyntaxError:
        return None
    assert tree is not None and isinstance(tree, ast.Module)
    if len(tree.body) == 1:
        stmt = tree.body[0]
        if isinstance(stmt, ast.Expr):
            expr = stmt.value
            if isinstance(expr, ast.Call):
                args = expr.args
                if pos < len(args):
                    res = args[pos]
                    if isinstance(res, ast.Name):
                        return res.id
    return None


def data_to_json(data, subsample_array: bool = True) -> Union[list, dict, int, float]:
    if type(data) is tuple:
        # convert to list
        return [data_to_json(elem, subsample_array) for elem in data]
    if type(data) is list:
        return [data_to_json(elem, subsample_array) for elem in data]
    elif type(data) is dict:
        return {key: data_to_json(data[key], subsample_array) for key in data}
    elif isinstance(data, np.ndarray):
        return ndarray_to_json(data, subsample_array)
    elif type(data) is scipy.sparse.csr_matrix:
        return ndarray_to_json(data.toarray(), subsample_array)
    elif isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        np_array = data.values
        return ndarray_to_json(np_array, subsample_array)
    elif torch_installed and isinstance(data, torch.Tensor):
        np_array = data.detach().numpy()
        return ndarray_to_json(np_array, subsample_array)
    elif (
        isinstance(data, np.int64)
        or isinstance(data, np.int32)
        or isinstance(data, np.int16)
    ):
        return int(data)
    elif isinstance(data, np.float64) or isinstance(data, np.float32):
        return float(data)
    else:
        return data


def is_empty_dict(val) -> bool:
    return isinstance(val, dict) and len(val) == 0


def dict_without(orig_dict: Dict[str, Any], key: str) -> Dict[str, Any]:
    return {k: orig_dict[k] for k in orig_dict if k != key}


def json_lookup(ptr, jsn, default=None):
    steps = ptr.split("/")
    sub_jsn = jsn
    for s in steps:
        if s not in sub_jsn:
            return default
        sub_jsn = sub_jsn[s]
    return sub_jsn


def ndarray_to_json(arr: np.ndarray, subsample_array: bool = True) -> Union[list, dict]:
    # sample 10 rows and no limit on columns
    num_subsamples: List[int]
    if subsample_array:
        num_subsamples = [10, np.iinfo(int).max, np.iinfo(int).max]
    else:
        num_subsamples = [
            np.iinfo(int).max,
            np.iinfo(int).max,
            np.iinfo(int).max,
        ]

    def subarray_to_json(indices: Tuple[int, ...]) -> Any:
        if len(indices) == len(arr.shape):
            if (
                isinstance(arr[indices], bool)
                or isinstance(arr[indices], int)
                or isinstance(arr[indices], float)
                or isinstance(arr[indices], str)
            ):
                return arr[indices]
            elif np.issubdtype(arr.dtype, np.bool_):
                return bool(arr[indices])
            elif np.issubdtype(arr.dtype, np.integer):
                return int(arr[indices])
            elif np.issubdtype(arr.dtype, np.number):
                return float(arr[indices])
            elif arr.dtype.kind in ["U", "S", "O"]:
                return str(arr[indices])
            else:
                raise ValueError(
                    f"Unexpected dtype {arr.dtype}, "
                    f"kind {arr.dtype.kind}, "
                    f"type {type(arr[indices])}."
                )
        else:
            assert len(indices) < len(arr.shape)
            return [
                subarray_to_json(indices + (i,))
                for i in range(
                    min(num_subsamples[len(indices)], arr.shape[len(indices)])
                )
            ]

    return subarray_to_json(())


def split_with_schemas(estimator, all_X, all_y, indices, train_indices=None):
    subset_X, subset_y = _safe_split(estimator, all_X, all_y, indices, train_indices)
    if hasattr(all_X, "json_schema"):
        n_rows = subset_X.shape[0]
        schema = {
            "type": "array",
            "minItems": n_rows,
            "maxItems": n_rows,
            "items": all_X.json_schema["items"],
        }
        lale.datasets.data_schemas.add_schema(subset_X, schema)
    if hasattr(all_y, "json_schema"):
        n_rows = subset_y.shape[0]
        schema = {
            "type": "array",
            "minItems": n_rows,
            "maxItems": n_rows,
            "items": all_y.json_schema["items"],
        }
        lale.datasets.data_schemas.add_schema(subset_y, schema)
    return subset_X, subset_y


def fold_schema(X, y, cv=1, is_classifier=True):
    def fold_schema_aux(data, n_rows):
        orig_schema = lale.datasets.data_schemas.to_schema(data)
        aux_result = {**orig_schema, "minItems": n_rows, "maxItems": n_rows}
        return aux_result

    n_splits = cv if isinstance(cv, int) else cv.get_n_splits()
    try:
        n_samples = X.shape[0] if hasattr(X, "shape") else len(X)
    except TypeError:  # raised for Spark dataframes.
        n_samples = X.count() if hasattr(X, "count") else 0

    if n_splits == 1:
        n_rows_fold = n_samples
    elif is_classifier:
        n_classes = len(set(y))
        n_rows_unstratified = (n_samples // n_splits) * (n_splits - 1)
        # in stratified case, fold sizes can differ by up to n_classes
        n_rows_fold = max(1, n_rows_unstratified - n_classes)
    else:
        n_rows_fold = (n_samples // n_splits) * (n_splits - 1)
    schema_X = fold_schema_aux(X, n_rows_fold)
    schema_y = fold_schema_aux(y, n_rows_fold)
    result = {"properties": {"X": schema_X, "y": schema_y}}
    return result


def cross_val_score_track_trials(
    estimator, X, y=None, scoring=accuracy_score, cv=5, args_to_scorer=None
):
    """
    Use the given estimator to perform fit and predict for splits defined by 'cv' and compute the given score on
    each of the splits.

    Parameters
    ----------

    estimator: A valid sklearn_wrapper estimator
    X, y: Valid data and target values that work with the estimator
    scoring: string or a scorer object created using
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer.
        A string from sklearn.metrics.SCORERS.keys() can be used or a scorer created from one of
        sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics).
        A completely custom scorer object can be created from a python function following the example at
        https://scikit-learn.org/stable/modules/model_evaluation.html
        The metric has to return a scalar value,
    cv: an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
        Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
        Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.
    args_to_scorer: A dictionary of additional keyword arguments to pass to the scorer.
                Used for cases where the scorer has a signature such as ``scorer(estimator, X, y, **kwargs)``.
    Returns
    -------
        cv_results: a list of scores corresponding to each cross validation fold
    """
    if isinstance(cv, int):
        cv = StratifiedKFold(cv)

    if args_to_scorer is None:
        args_to_scorer = {}
    scorer = check_scoring(estimator, scoring=scoring)
    cv_results: List[float] = []
    log_loss_results = []
    time_results = []
    for train, test in cv.split(X, y):
        X_train, y_train = split_with_schemas(estimator, X, y, train)
        X_test, y_test = split_with_schemas(estimator, X, y, test, train)
        start = time.time()
        # Not calling sklearn.base.clone() here, because:
        #  (1) For Lale pipelines, clone() calls the pipeline constructor
        #      with edges=None, so the resulting topology is incorrect.
        #  (2) For Lale individual operators, the fit() method already
        #      clones the impl object, so cloning again is redundant.
        trained = estimator.fit(X_train, y_train)
        score_value = scorer(trained, X_test, y_test, **args_to_scorer)
        execution_time = time.time() - start
        # not all estimators have predict probability
        try:
            y_pred_proba = trained.predict_proba(X_test)
            logloss = log_loss(y_true=y_test, y_pred=y_pred_proba)
            log_loss_results.append(logloss)
        except BaseException:
            logger.debug("Warning, log loss cannot be computed")
        cv_results.append(score_value)
        time_results.append(execution_time)
    result = (
        np.array(cv_results).mean(),
        np.array(log_loss_results).mean(),
        np.array(time_results).mean(),
    )
    return result


def cross_val_score(estimator, X, y=None, scoring=accuracy_score, cv=5):
    """
    Use the given estimator to perform fit and predict for splits defined by 'cv' and compute the given score on
    each of the splits.

    Parameters
    ----------

    estimator: A valid sklearn_wrapper estimator
    X, y: Valid data and target values that work with the estimator
    scoring: a scorer object from sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics)
        Default value is accuracy_score.
    cv: an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
        Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
        Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.

    Returns
    -------
    cv_results: a list of scores corresponding to each cross validation fold
    """
    if isinstance(cv, int):
        cv = StratifiedKFold(cv)

    cv_results = []
    for train, test in cv.split(X, y):
        X_train, y_train = split_with_schemas(estimator, X, y, train)
        X_test, y_test = split_with_schemas(estimator, X, y, test, train)
        trained_estimator = estimator.fit(X_train, y_train)
        predicted_values = trained_estimator.predict(X_test)
        cv_results.append(scoring(y_test, predicted_values))

    return cv_results


def create_individual_op_using_reflection(class_name, operator_name, param_dict):
    instance = None
    if class_name is not None:
        class_name_parts = class_name.split(".")
        assert (
            len(class_name_parts)
        ) > 1, (
            "The class name needs to be fully qualified, i.e. module name + class name"
        )
        module_name = ".".join(class_name_parts[0:-1])
        class_name = class_name_parts[-1]

        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)

        if param_dict is None:
            instance = class_()
        else:
            instance = class_(**param_dict)
    return instance


if TYPE_CHECKING:
    import lale.operators


def to_graphviz(
    lale_operator: "lale.operators.Operator",
    ipython_display: bool = True,
    call_depth: int = 1,
    **dot_graph_attr,
):
    import lale.json_operator
    import lale.operators
    import lale.visualize

    if not isinstance(lale_operator, lale.operators.Operator):
        raise TypeError("The input to to_graphviz needs to be a valid LALE operator.")
    jsn = lale.json_operator.to_json(lale_operator, call_depth=call_depth + 1)
    dot = lale.visualize.json_to_graphviz(jsn, ipython_display, dot_graph_attr)
    return dot


def println_pos(message, out_file=sys.stdout):
    tb = traceback.extract_stack()[-2]
    match = re.search(r"<ipython-input-([0-9]+)-", tb[0])
    if match:
        pos = "notebook cell [{}] line {}".format(match[1], tb[1])
    else:
        pos = "{}:{}".format(tb[0], tb[1])
    strtime = time.strftime("%Y-%m-%d_%H-%M-%S")
    to_log = "{}: {} {}".format(pos, strtime, message)
    print(to_log, file=out_file)
    if match:
        os.system("echo {}".format(to_log))


def instantiate_from_hyperopt_search_space(obj_hyperparams, new_hyperparams):
    if isinstance(new_hyperparams, dict) and LALE_NESTED_SPACE_KEY in new_hyperparams:
        sub_params = new_hyperparams[LALE_NESTED_SPACE_KEY]

        sub_op = obj_hyperparams
        if isinstance(sub_op, list):
            if len(sub_op) == 1:
                sub_op = sub_op[0]
            else:
                step_index, step_params = list(sub_params)[0]
                if step_index < len(sub_op):
                    sub_op = sub_op[step_index]
                    sub_params = step_params

        return create_instance_from_hyperopt_search_space(sub_op, sub_params)

    elif isinstance(new_hyperparams, (list, tuple)):
        assert isinstance(obj_hyperparams, (list, tuple))
        params_len = len(new_hyperparams)
        assert params_len == len(obj_hyperparams)
        res: Optional[List[Any]] = None

        for i in range(params_len):
            nhi = new_hyperparams[i]
            ohi = obj_hyperparams[i]
            updated_params = instantiate_from_hyperopt_search_space(ohi, nhi)
            if updated_params is not None:
                if res is None:
                    res = list(new_hyperparams)
                res[i] = updated_params
        if res is not None:
            if isinstance(obj_hyperparams, tuple):
                return tuple(res)
            else:
                return res
        # workaround for what seems to be a hyperopt bug
        # where hyperopt returns a tuple even though the
        # hyperopt search space specifies a list
        is_obj_tuple = isinstance(obj_hyperparams, tuple)
        is_new_tuple = isinstance(new_hyperparams, tuple)
        if is_obj_tuple != is_new_tuple:
            if is_obj_tuple:
                return tuple(new_hyperparams)
            else:
                return list(new_hyperparams)
        return None

    elif isinstance(new_hyperparams, dict):
        assert isinstance(obj_hyperparams, dict)

        for k, sub_params in new_hyperparams.items():
            if k in obj_hyperparams:
                sub_op = obj_hyperparams[k]
                updated_params = instantiate_from_hyperopt_search_space(
                    sub_op, sub_params
                )
                if updated_params is not None:
                    new_hyperparams[k] = updated_params
        return None
    else:
        return None


def create_instance_from_hyperopt_search_space(
    lale_object, hyperparams
) -> "lale.operators.Operator":
    """
    Hyperparams is a n-tuple of dictionaries of hyper-parameters, each
    dictionary corresponds to an operator in the pipeline
    """
    # lale_object can either be an individual operator, a pipeline or an operatorchoice
    # Validate that the number of elements in the n-tuple is the same
    # as the number of steps in the current pipeline

    from lale.operators import (
        BasePipeline,
        OperatorChoice,
        PlannedIndividualOp,
        TrainableOperator,
        TrainablePipeline,
    )

    if isinstance(lale_object, PlannedIndividualOp):
        new_hyperparams: Dict[str, Any] = dict_without(hyperparams, "name")
        hps = lale_object.hyperparams()
        if hps:
            obj_hyperparams = dict(hps)
        else:
            obj_hyperparams = {}

        for k, sub_params in new_hyperparams.items():
            if k in obj_hyperparams:
                sub_op = obj_hyperparams[k]
                updated_params = instantiate_from_hyperopt_search_space(
                    sub_op, sub_params
                )
                if updated_params is not None:
                    new_hyperparams[k] = updated_params

        all_hyperparams = {**obj_hyperparams, **new_hyperparams}
        return lale_object(**all_hyperparams)
    elif isinstance(lale_object, BasePipeline):
        steps = lale_object.steps_list()
        if len(hyperparams) != len(steps):
            raise ValueError(
                "The number of steps in the hyper-parameter space does not match the number of steps in the pipeline."
            )
        op_instances = []
        edges = lale_object.edges()
        # op_map:Dict[PlannedOpType, TrainableOperator] = {}
        op_map = {}
        for op_index, sub_params in enumerate(hyperparams):
            sub_op = steps[op_index]
            op_instance = create_instance_from_hyperopt_search_space(sub_op, sub_params)
            assert isinstance(op_instance, TrainableOperator)
            assert (
                isinstance(sub_op, OperatorChoice)
                or sub_op.class_name() == op_instance.class_name()
            ), f"sub_op {sub_op.class_name()}, op_instance {op_instance.class_name()}"
            op_instances.append(op_instance)
            op_map[sub_op] = op_instance

        # trainable_edges:List[Tuple[TrainableOperator, TrainableOperator]]
        try:
            trainable_edges = [(op_map[x], op_map[y]) for (x, y) in edges]
        except KeyError as e:
            raise ValueError(
                "An edge was found with an endpoint that is not a step (" + str(e) + ")"
            )

        return TrainablePipeline(op_instances, trainable_edges, ordered=True)  # type: ignore
    elif isinstance(lale_object, OperatorChoice):
        # Hyperopt search space for an OperatorChoice is generated as a dictionary with a single element
        # corresponding to the choice made, the only key is the index of the step and the value is
        # the params corresponding to that step.
        step_index: int
        choices = lale_object.steps_list()

        if len(choices) == 1:
            step_index = 0
        else:
            step_index_str, hyperparams = list(hyperparams.items())[0]
            step_index = int(step_index_str)
        step_object = choices[step_index]
        return create_instance_from_hyperopt_search_space(step_object, hyperparams)
    else:
        assert False, f"Unknown operator type: {type(lale_object)}"


def import_from_sklearn_pipeline(sklearn_pipeline, fitted=True, is_hyperparam=False):
    # For all pipeline steps, identify equivalent lale wrappers if present,
    # if not, call make operator on sklearn classes and create a lale pipeline.
    # For higher order operators, we allow hyperparameters to be trainable even with
    # fitted is True. This is achieved using the is_hyperparam flag.

    def find_lale_wrapper(sklearn_obj):
        module_names = [
            "lale.lib.sklearn",
            "lale.lib.autoai_libs",
            "lale.lib.xgboost",
            "lale.lib.lightgbm",
            "lale.lib.snapml",
        ]

        try:
            import autoai_ts_libs  # type: ignore # noqa

            module_names.append("lale.lib.autoai_ts_libs")
        except ImportError:
            pass

        lale_wrapper_found = False
        class_name = sklearn_obj.__class__.__name__
        for module_name in module_names:
            try:
                module = importlib.import_module(module_name)
            except ModuleNotFoundError:
                continue
            try:
                class_ = getattr(module, class_name)
                lale_wrapper_found = True
                break
            except AttributeError:
                continue
        else:
            return lale_wrapper_found, sklearn_obj
        return lale_wrapper_found, class_

    import lale.operators
    import lale.type_checking

    sklearn_obj = sklearn_pipeline

    if isinstance(sklearn_obj, lale.operators.TrainableIndividualOp) and fitted:
        if hasattr(sklearn_obj, "_trained"):
            return sklearn_obj._trained
        elif is_hyperparam or not hasattr(
            sklearn_obj._impl_instance(), "fit"
        ):  # Operators such as NoOp do not have a fit, so return them as is.
            return sklearn_obj
        else:
            raise ValueError(
                f"""The input pipeline has an operator {sklearn_obj} that is not trained and fitted is set to True,
                please pass fitted=False if you want a trainable pipeline as output."""
            )
    elif isinstance(sklearn_obj, lale.operators.Operator):
        return sklearn_obj

    if isinstance(sklearn_pipeline, sklearn.pipeline.Pipeline):
        nested_pipeline_steps = sklearn_pipeline.named_steps
        nested_pipeline_lale_named_steps = [
            (
                nested_pipeline_step[0],
                import_from_sklearn_pipeline(
                    nested_pipeline_step[1], fitted=fitted, is_hyperparam=is_hyperparam
                ),
            )
            for nested_pipeline_step in nested_pipeline_steps.items()
        ]
        if type(sklearn_pipeline) == sklearn.pipeline.Pipeline:
            nested_pipeline_lale_objects = [
                nested_pipeline_lale_named_step[1]
                for nested_pipeline_lale_named_step in nested_pipeline_lale_named_steps
            ]
            lale_op_obj = lale.operators.make_pipeline(*nested_pipeline_lale_objects)
        else:
            lale_wrapper_found, wrapper_class = find_lale_wrapper(sklearn_pipeline)
            if lale_wrapper_found:
                # This is a custom subclass of sklearn pipeline, so use the wrapper class
                # instead of creating a lale pipeline
                # We assume it has a hyperparameter `steps`.
                if (
                    not fitted
                ):  # If fitted is False, we do not want to return a Trained operator.
                    lale_op = wrapper_class
                else:
                    lale_op = lale.operators.TrainedIndividualOp(
                        wrapper_class._name,
                        wrapper_class._impl,
                        wrapper_class._schemas,
                        None,
                        _lale_trained=True,
                    )
                lale_op_obj = lale_op(steps=nested_pipeline_lale_named_steps)
            else:  # no conversion to lale if a wrapper is not found for a subclass of pipeline
                return sklearn_pipeline
    elif isinstance(sklearn_pipeline, sklearn.pipeline.FeatureUnion):
        transformer_list = sklearn_pipeline.transformer_list
        concat_predecessors = [
            import_from_sklearn_pipeline(
                transformer[1], fitted=fitted, is_hyperparam=is_hyperparam
            )
            for transformer in transformer_list
        ]
        lale_op_obj = lale.operators.make_union(*concat_predecessors)
    else:
        # Validate that the sklearn_obj is a valid sklearn-compatible object
        if sklearn_obj is None or not hasattr(sklearn_obj, "get_params"):
            raise ValueError(
                f"The input pipeline has a step {sklearn_obj} that is not scikit-learn compatible."
            )
        orig_hyperparams = sklearn_obj.get_params(deep=False)
        higher_order = False
        for hp_name, hp_val in orig_hyperparams.items():
            higher_order = higher_order or hasattr(hp_val, "get_params")
        if higher_order:
            hyperparams = {}
            for hp_name, hp_val in orig_hyperparams.items():
                if hasattr(hp_val, "get_params"):
                    nested_op = import_from_sklearn_pipeline(
                        hp_val, fitted, is_hyperparam=True
                    )  # allow nested_op to be trainable
                    hyperparams[hp_name] = nested_op
                else:
                    hyperparams[hp_name] = hp_val
        else:
            hyperparams = orig_hyperparams

        lale_wrapper_found, class_ = find_lale_wrapper(sklearn_obj)
        if not lale_wrapper_found:
            return class_  # Return the original object

        if (
            not fitted
        ):  # If fitted is False, we do not want to return a Trained operator.
            lale_op = class_
        else:
            lale_op = lale.operators.TrainedIndividualOp(
                class_._name, class_._impl, class_._schemas, None, _lale_trained=True
            )
        class_ = lale_op(**hyperparams)
        lale_op_obj = class_
        if lale_wrapper_found and hasattr(class_._impl_instance(), "_wrapped_model"):
            wrapped_model = copy.deepcopy(sklearn_obj)
            class_._impl_instance()._wrapped_model = wrapped_model
        else:  # If there is no lale wrapper, there is no _wrapped_model
            class_._impl = copy.deepcopy(sklearn_obj)
            class_._impl_class_ = class_._impl.__class__
            lale_op_obj = class_
    return lale_op_obj


class val_wrapper:
    """This is used to wrap values that cause problems for hyper-optimizer backends
    lale will unwrap these when given them as the value of a hyper-parameter"""

    def __init__(self, base):
        self._base = base

    def unwrap_self(self):
        return self._base

    @classmethod
    def unwrap(cls, obj):
        if isinstance(obj, cls):
            return cls.unwrap(obj.unwrap_self())
        else:
            return obj


def append_batch(data, batch_data):
    if data is None:
        return batch_data
    elif isinstance(data, np.ndarray):
        if isinstance(batch_data, np.ndarray):
            if len(data.shape) == 1 and len(batch_data.shape) == 1:
                return np.concatenate([data, batch_data])
            else:
                return np.vstack((data, batch_data))
    elif isinstance(data, tuple):
        X, y = data
        if isinstance(batch_data, tuple):
            batch_X, batch_y = batch_data
            X = append_batch(X, batch_X)
            y = append_batch(y, batch_y)
            return X, y
    elif torch_installed and isinstance(data, torch.Tensor):
        if isinstance(batch_data, torch.Tensor):
            return torch.cat((data, batch_data))
    try:
        import h5py

        if isinstance(data, h5py.File):
            if isinstance(batch_data, tuple):
                batch_X, batch_y = batch_data
    except ModuleNotFoundError:
        pass

    # TODO:Handle dataframes


def create_data_loader(X, y=None, batch_size=1, num_workers=0):
    """A function that takes a dataset as input and outputs a Pytorch dataloader.

    Parameters
    ----------
    X : Input data.
        The formats supported are Pandas DataFrame, Numpy array,
        a sparse matrix, torch.tensor, torch.utils.data.Dataset, path to a HDF5 file,
        lale.util.batch_data_dictionary_dataset.BatchDataDict,
        a Python dictionary of the format `{"dataset": torch.utils.data.Dataset,
        "collate_fn":collate_fn for torch.utils.data.DataLoader}`
    y : Labels., optional
        Supported formats are Numpy array or Pandas series, by default None
    batch_size : int, optional
        Number of samples in each batch, by default 1
    num_workers : int, optional
        Number of workers used by the data loader, by default 0

    Returns
    -------
    torch.utils.data.DataLoader

    Raises
    ------
    TypeError
        Raises a TypeError if the input format is not supported.
    """
    import torch
    from torch.utils.data import DataLoader, Dataset, TensorDataset

    from lale.util.batch_data_dictionary_dataset import BatchDataDict
    from lale.util.hdf5_to_torch_dataset import HDF5TorchDataset
    from lale.util.numpy_to_torch_dataset import NumpyTorchDataset

    collate_fn = None
    worker_init_fn = None

    def numpy_collate_fn(batch):
        return_X = None
        return_y = None
        for item in batch:
            if isinstance(item, tuple):
                if return_X is None:
                    return_X = item[0]
                else:
                    return_X = np.vstack((return_X, item[0]))
                if return_y is None:
                    return_y = item[1]
                else:
                    return_y = np.vstack((return_y, item[1]))
            else:
                if return_X is None:
                    return_X = item
                else:
                    return_X = np.vstack((return_X, item))
        if return_y is not None:
            if len(return_y.shape) > 1 and return_y.shape[1] == 1:
                return_y = np.reshape(return_y, (len(return_y),))
            return return_X, return_y
        else:
            return return_X

    if isinstance(X, Dataset):
        dataset = X
    elif isinstance(X, pd.DataFrame):
        X = X.to_numpy()
        if isinstance(y, pd.Series):
            y = y.to_numpy()
        dataset = NumpyTorchDataset(X, y)
        collate_fn = numpy_collate_fn
    elif isinstance(X, scipy.sparse.csr.csr_matrix):
        # unfortunately, NumpyTorchDataset won't accept a subclass of np.ndarray
        X = X.toarray()
        if isinstance(y, lale.datasets.data_schemas.NDArrayWithSchema):
            y = y.view(np.ndarray)
        dataset = NumpyTorchDataset(X, y)
        collate_fn = numpy_collate_fn
    elif isinstance(X, np.ndarray):
        # unfortunately, NumpyTorchDataset won't accept a subclass of np.ndarray
        if isinstance(X, lale.datasets.data_schemas.NDArrayWithSchema):
            X = X.view(np.ndarray)
        if isinstance(y, lale.datasets.data_schemas.NDArrayWithSchema):
            y = y.view(np.ndarray)
        dataset = NumpyTorchDataset(X, y)
        collate_fn = numpy_collate_fn
    elif isinstance(X, str):  # Assume that this is path to hdf5 file
        dataset = HDF5TorchDataset(X)
    elif isinstance(X, BatchDataDict):
        dataset = X

        def my_collate_fn(batch):
            return batch[
                0
            ]  # because BatchDataDict's get_item returns a batch, so no collate is required.

        return DataLoader(dataset, batch_size=1, collate_fn=my_collate_fn)
    elif isinstance(X, dict):  # Assumed that it is data indexed by batch number
        if "dataset" in X:
            dataset = X["dataset"]
            collate_fn = X.get("collate_fn", None)
            worker_init_fn = getattr(dataset, "worker_init_fn", None)
        else:
            return [X]
    elif isinstance(X, torch.Tensor) and y is not None:
        if isinstance(y, np.ndarray):
            y = torch.from_numpy(y)
        dataset = TensorDataset(X, y)
    elif isinstance(X, torch.Tensor):
        dataset = TensorDataset(X)
    else:
        raise TypeError(
            "Can not create a data loader for a dataset with type {}".format(type(X))
        )
    return DataLoader(
        dataset,
        batch_size=batch_size,
        collate_fn=collate_fn,
        num_workers=num_workers,
        worker_init_fn=worker_init_fn,
    )


def write_batch_output_to_file(
    file_obj,
    file_path,
    total_len,
    batch_idx,
    batch_X,
    batch_y,
    batch_out_X,
    batch_out_y,
):
    if file_obj is None and file_path is None:
        raise ValueError("Only one of the file object or file path can be None.")
    if file_obj is None:
        import h5py

        file_obj = h5py.File(file_path, "w")
        # estimate the size of the dataset based on the first batch output size
        transform_ratio = int(len(batch_out_X) / len(batch_X))
        if len(batch_out_X.shape) == 1:
            h5_data_shape = (transform_ratio * total_len,)
        elif len(batch_out_X.shape) == 2:
            h5_data_shape = (transform_ratio * total_len, batch_out_X.shape[1])
        elif len(batch_out_X.shape) == 3:
            h5_data_shape = (
                transform_ratio * total_len,
                batch_out_X.shape[1],
                batch_out_X.shape[2],
            )
        else:
            raise ValueError(
                "batch_out_X is expected to be a 1-d, 2-d or 3-d array. Any other data types are not handled."
            )
        dataset = file_obj.create_dataset(
            name="X", shape=h5_data_shape, chunks=True, compression="gzip"
        )
        if batch_out_y is None and batch_y is not None:
            batch_out_y = batch_y
        if batch_out_y is not None:
            if len(batch_out_y.shape) == 1:
                h5_labels_shape = (transform_ratio * total_len,)
            elif len(batch_out_y.shape) == 2:
                h5_labels_shape = (transform_ratio * total_len, batch_out_y.shape[1])
            else:
                raise ValueError(
                    "batch_out_y is expected to be a 1-d or 2-d array. Any other data types are not handled."
                )
            dataset = file_obj.create_dataset(
                name="y", shape=h5_labels_shape, chunks=True, compression="gzip"
            )
    dataset = file_obj["X"]
    dataset[
        batch_idx * len(batch_out_X) : (batch_idx + 1) * len(batch_out_X)
    ] = batch_out_X
    if batch_out_y is not None or batch_y is not None:
        labels = file_obj["y"]
        if batch_out_y is not None:
            labels[
                batch_idx * len(batch_out_y) : (batch_idx + 1) * len(batch_out_y)
            ] = batch_out_y
        else:
            labels[batch_idx * len(batch_y) : (batch_idx + 1) * len(batch_y)] = batch_y
    return file_obj


def add_missing_values(orig_X, missing_rate=0.1, seed=None):
    # see scikit-learn.org/stable/auto_examples/impute/plot_missing_values.html
    n_samples, n_features = orig_X.shape
    n_missing_samples = int(n_samples * missing_rate)
    if seed is None:
        rng = np.random.RandomState()
    else:
        rng = np.random.RandomState(seed)
    missing_samples = np.zeros(n_samples, dtype=bool)
    missing_samples[:n_missing_samples] = True
    rng.shuffle(missing_samples)
    missing_features = rng.randint(0, n_features, n_missing_samples)
    missing_X = orig_X.copy()
    if isinstance(missing_X, np.ndarray):
        missing_X[missing_samples, missing_features] = np.nan
    else:
        assert isinstance(missing_X, pd.DataFrame)
        i_missing_sample = 0
        for i_sample in range(n_samples):
            if missing_samples[i_sample]:
                i_feature = missing_features[i_missing_sample]
                i_missing_sample += 1
                missing_X.iloc[i_sample, i_feature] = np.nan
    return missing_X


# helpers for manipulating (extended) sklearn style paths.
# documentation of the path format is part of the operators module docstring


def partition_sklearn_params(
    d: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
    sub_parts: Dict[str, Dict[str, Any]] = {}
    main_parts: Dict[str, Any] = {}

    for k, v in d.items():
        ks = k.split("__", 1)
        if len(ks) == 1:
            assert k not in main_parts
            main_parts[k] = v
        else:
            assert len(ks) == 2
            bucket: Dict[str, Any] = {}
            group: str = ks[0]
            param: str = ks[1]
            if group in sub_parts:
                bucket = sub_parts[group]
            else:
                sub_parts[group] = bucket
            assert param not in bucket
            bucket[param] = v
    return (main_parts, sub_parts)


def partition_sklearn_choice_params(d: Dict[str, Any]) -> Tuple[int, Dict[str, Any]]:
    discriminant_value: int = -1
    choice_parts: Dict[str, Any] = {}

    for k, v in d.items():
        if k == discriminant_name:
            assert discriminant_value == -1
            discriminant_value = int(v)
        else:
            k_rest = unnest_choice(k)
            choice_parts[k_rest] = v
    assert discriminant_value != -1
    return (discriminant_value, choice_parts)


DUMMY_SEARCH_SPACE_GRID_PARAM_NAME: str = "$"
discriminant_name: str = "?"
choice_prefix: str = "?"
structure_type_name: str = "#"
structure_type_list: str = "list"
structure_type_tuple: str = "tuple"
structure_type_dict: str = "dict"


def get_name_and_index(name: str) -> Tuple[str, int]:
    """given a name of the form "name@i", returns (name, i)
    if given a name of the form "name", returns (name, 0)
    """
    splits = name.split("@", 1)
    if len(splits) == 1:
        return splits[0], 0
    else:
        return splits[0], int(splits[1])


def make_degen_indexed_name(name, index):
    return f"{name}@{index}"


def make_indexed_name(name, index):
    if index == 0:
        return name
    else:
        return f"{name}@{index}"


def make_array_index_name(index, is_tuple: bool = False):
    sep = "##" if is_tuple else "#"
    return f"{sep}{str(index)}"


def is_numeric_structure(structure_type: str):

    if structure_type == "list" or structure_type == "tuple":
        return True
    elif structure_type == "dict":
        return False
    else:
        assert False, f"Unknown structure type {structure_type} found"


V = TypeVar("V")


def nest_HPparam(name: str, key: str):
    if key == DUMMY_SEARCH_SPACE_GRID_PARAM_NAME:
        # we can get rid of the dummy now, since we have a name for it
        return name
    return name + "__" + key


def nest_HPparams(name: str, grid: Mapping[str, V]) -> Dict[str, V]:
    return {(nest_HPparam(name, k)): v for k, v in grid.items()}


def nest_all_HPparams(
    name: str, grids: Iterable[Mapping[str, V]]
) -> List[Dict[str, V]]:
    """Given the name of an operator in a pipeline, this transforms every key(parameter name) in the grids
    to use the operator name as a prefix (separated by __).  This is the convention in scikit-learn pipelines.
    """
    return [nest_HPparams(name, grid) for grid in grids]


def nest_choice_HPparam(key: str):
    return choice_prefix + key


def nest_choice_HPparams(grid: Mapping[str, V]) -> Dict[str, V]:
    return {(nest_choice_HPparam(k)): v for k, v in grid.items()}


def nest_choice_all_HPparams(grids: Iterable[Mapping[str, V]]) -> List[Dict[str, V]]:
    """this transforms every key(parameter name) in the grids
    to be nested under a choice, using a ? as a prefix (separated by __).  This is the convention in scikit-learn pipelines.
    """
    return [nest_choice_HPparams(grid) for grid in grids]


def unnest_choice(k: str) -> str:
    assert k.startswith(choice_prefix)
    return k[len(choice_prefix) :]


def unnest_HPparams(k: str) -> List[str]:
    return k.split("__")


def are_hyperparameters_equal(hyperparam1, hyperparam2):
    if isinstance(
        hyperparam1, np.ndarray
    ):  # hyperparam2 is from schema default, so it may not always be an array
        return np.all(hyperparam1 == hyperparam2)
    else:
        return hyperparam1 == hyperparam2


def _is_ast_subscript(expr):
    return isinstance(expr, ast.Subscript)


def _is_ast_attribute(expr):
    return isinstance(expr, ast.Attribute)


def _is_ast_constant(expr):
    return isinstance(expr, ast.Constant)


def _is_ast_subs_or_attr(expr):
    return isinstance(expr, ast.Subscript) or isinstance(expr, ast.Attribute)


def _is_ast_call(expr):
    return isinstance(expr, ast.Call)


def _is_ast_name(expr):
    return isinstance(expr, ast.Name)


def _ast_func_id(expr):
    if isinstance(expr, ast.Name):
        return expr.id
    else:
        raise ValueError("function name expected")


def _is_df(d):
    return isinstance(d, pd.DataFrame) or isinstance(d, spark_df)


def _is_pandas_df(df):
    return isinstance(df, pd.DataFrame)


def _is_spark_df(df):
    if spark_installed:
        return isinstance(df, spark_df)


def _ensure_pandas(df):
    if _is_spark_df(df):
        return df.toPandas()
    assert _is_pandas_df, type(df)
    return df


def _get_subscript_value(subscript_expr):
    if isinstance(subscript_expr.slice, ast.Constant):  # for Python 3.9
        subscript_value = subscript_expr.slice.value
    else:
        subscript_value = subscript_expr.slice.value.s  # type: ignore
    return subscript_value



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import os
import sys
import warnings

import numpy as np
from scipy.signal import resample
from sklearn import preprocessing

import lale.docstrings
import lale.operators

sys.path.append(os.getcwd())


seizure_type_data = collections.namedtuple(
    "seizure_type_data", ["seizure_type", "data"]
)

# Some of the classes and modules have been taken from https://github.com/MichaelHills/seizure-detection


class Slice:
    """
    Job: Take a slice of the data on the last axis.
    Note: Slice(x, y) works like a normal python slice, that is x to (y-1) will be taken.
    """

    def __init__(self, start, stop):
        self.start = start
        self.stop = stop + 1

    def get_name(self):
        return "slice%d-%d" % (self.start, self.stop)

    def apply(self, data):
        s = [
            slice(None),
        ] * data.ndim
        s[-1] = slice(self.start, self.stop)
        return data[s]


class Magnitude:
    """
    Job: Take magnitudes of Complex data
    """

    def get_name(self):
        return "mag"

    def apply(self, data):
        return np.absolute(data)


class Log10:
    """
    Apply Log10
    """

    def get_name(self):
        return "log10"

    def apply(self, data):
        # 10.0 * log10(re * re + im * im)
        indices = np.where(data <= 0)
        data[indices] = np.max(data)
        data[indices] = np.min(data) * 0.1
        return np.log10(data)


class Pipeline(object):
    """
    A Pipeline is an object representing the data transformations to make
    on the input data, finally outputting extracted features.
    pipeline: List of transforms to apply one by one to the input data
    """

    def __init__(self, pipeline):
        self.transforms = pipeline
        names = [t.get_name() for t in self.transforms]
        self.name = "empty" if len(names) == 0 else "_".join(names)

    def get_name(self):
        return self.name

    def apply(self, data):
        for transform in self.transforms:
            data = transform.apply(data)
        return data


class FFT:
    """
    Apply Fast Fourier Transform to the last axis.
    """

    def get_name(self):
        return "fft"

    def apply(self, data):
        axis = data.ndim - 1
        return np.fft.rfft(data, axis=axis)


class Resample:
    """
    Resample time-series data.
    """

    def __init__(self, sample_rate):
        self.f = sample_rate

    def get_name(self):
        return "resample%d" % self.f

    def apply(self, data):
        axis = data.ndim - 1
        if data.shape[-1] > self.f:
            return resample(data, self.f, axis=axis)
        return data


class StandardizeLast:
    """
    Scale across the last axis.
    """

    def get_name(self):
        return "standardize-last"

    def apply(self, data):
        return preprocessing.scale(data, axis=data.ndim - 1)


class StandardizeFirst:
    """
    Scale across the first axis.
    """

    def get_name(self):
        return "standardize-first"

    def apply(self, data):
        return preprocessing.scale(data, axis=0)


class CorrelationMatrix:
    """
    Calculate correlation coefficients matrix across all EEG channels.
    """

    def get_name(self):
        return "correlation-matrix"

    def apply(self, data):
        return np.corrcoef(data)


class Eigenvalues:
    """
    Take eigenvalues of a matrix, and sort them by magnitude in order to
    make them useful as features (as they have no inherent order).
    """

    def get_name(self):
        return "eigenvalues"

    def apply(self, data):
        w, v = np.linalg.eig(data)
        w = np.absolute(w)
        w.sort()
        return w


# Take the upper right triangle of a matrix
def upper_right_triangle(matrix):
    accum = []
    for i in range(matrix.shape[0]):
        for j in range(i + 1, matrix.shape[1]):
            accum.append(matrix[i, j])

    return np.array(accum)


class FreqCorrelation:
    """
    Correlation in the frequency domain. First take FFT with (start, end) slice options,
    then calculate correlation co-efficients on the FFT output, followed by calculating
    eigenvalues on the correlation co-efficients matrix.
    The output features are (fft, upper_right_diagonal(correlation_coefficients), eigenvalues)
    Features can be selected/omitted using the constructor arguments.
    """

    def __init__(
        self, start, end, scale_option, with_fft=False, with_corr=True, with_eigen=True
    ):
        self.start = start
        self.end = end
        self.scale_option = scale_option
        self.with_fft = with_fft
        self.with_corr = with_corr
        self.with_eigen = with_eigen
        assert scale_option in ("first_axis", "last_axis", "none")
        assert with_corr or with_eigen

    def get_name(self):
        selections = []
        if not self.with_corr:
            selections.append("nocorr")
        if not self.with_eigen:
            selections.append("noeig")
        if len(selections) > 0:
            selection_str = "-" + "-".join(selections)
        else:
            selection_str = ""
        return "freq-correlation-%d-%d-%s-%s%s" % (
            self.start,
            self.end,
            "withfft" if self.with_fft else "nofft",
            self.scale_option,
            selection_str,
        )

    def apply(self, data):
        data1 = FFT().apply(data)
        data1 = Slice(self.start, self.end).apply(data1)
        data1 = Magnitude().apply(data1)
        data1 = Log10().apply(data1)

        data2 = data1
        if self.scale_option == "first_axis":
            data2 = StandardizeFirst().apply(data2)
        elif self.scale_option == "last_axis":
            data2 = StandardizeLast().apply(data2)

        data2 = CorrelationMatrix().apply(data2)

        out = []

        if self.with_corr:
            ur = upper_right_triangle(data2)
            out.append(ur)
        if self.with_eigen:
            w = Eigenvalues().apply(data2)
            out.append(w)
        if self.with_fft:
            data1 = data1.ravel()
            out.append(data1)
        for d in out:
            assert d.ndim == 1

        return np.concatenate(out, axis=0)


class TimeCorrelation:
    """
    Correlation in the time domain. First downsample the data, then calculate correlation co-efficients
    followed by calculating eigenvalues on the correlation co-efficients matrix.
    The output features are (upper_right_diagonal(correlation_coefficients), eigenvalues)
    Features can be selected/omitted using the constructor arguments.
    """

    def __init__(self, max_hz, scale_option, with_corr=True, with_eigen=True):
        self.max_hz = max_hz
        self.scale_option = scale_option
        self.with_corr = with_corr
        self.with_eigen = with_eigen
        assert scale_option in ("first_axis", "last_axis", "none")
        assert with_corr or with_eigen

    def get_name(self):
        selections = []
        if not self.with_corr:
            selections.append("nocorr")
        if not self.with_eigen:
            selections.append("noeig")
        if len(selections) > 0:
            selection_str = "-" + "-".join(selections)
        else:
            selection_str = ""
        return "time-correlation-r%d-%s%s" % (
            self.max_hz,
            self.scale_option,
            selection_str,
        )

    def apply(self, data):
        # so that correlation matrix calculation doesn't crash
        for ch in data:
            if np.all(ch == 0.0):
                ch[-1] += 0.00001

        data1 = data
        if data1.shape[1] > self.max_hz:
            data1 = Resample(self.max_hz).apply(data1)

        if self.scale_option == "first_axis":
            data1 = StandardizeFirst().apply(data1)
        elif self.scale_option == "last_axis":
            data1 = StandardizeLast().apply(data1)

        data1 = CorrelationMatrix().apply(data1)

        out = []
        if self.with_corr:
            ur = upper_right_triangle(data1)
            out.append(ur)
        if self.with_eigen:
            w = Eigenvalues().apply(data1)
            out.append(w)

        for d in out:
            assert d.ndim == 1

        return np.concatenate(out, axis=0)


class FFTWithTimeFreqCorrelation:
    """
    Combines FFT with time and frequency correlation, taking both correlation coefficients and eigenvalues.
    """

    def __init__(self, start, end, max_hz, scale_option):
        self.start = start
        self.end = end
        self.max_hz = max_hz
        self.scale_option = scale_option
        assert scale_option in ("first_axis", "last_axis", "none")

    def get_name(self):
        return "fft-with-time-freq-corr-%d-%d-r%d-%s" % (
            self.start,
            self.end,
            self.max_hz,
            self.scale_option,
        )

    def apply(self, data):
        data1 = TimeCorrelation(self.max_hz, self.scale_option).apply(data)
        data2 = FreqCorrelation(
            self.start, self.end, self.scale_option, with_fft=True
        ).apply(data)
        assert data1.ndim == data2.ndim

        return np.concatenate((data1, data2), axis=data1.ndim - 1)


class _TimeFreqEigenVectorsImpl:
    def __init__(
        self,
        window_length=1,
        window_step=0.5,
        fft_min_freq=1,
        fft_max_freq=24,
        sampling_frequency=250,
    ):
        self.window_length = window_length
        self.window_step = window_step
        self.fft_min_freq = fft_min_freq
        self.fft_max_freq = fft_max_freq
        self.sampling_frequency = sampling_frequency

    def transform(self, X, y=None):
        warnings.filterwarnings("ignore")
        pipeline = Pipeline(
            [
                FFTWithTimeFreqCorrelation(
                    self.fft_min_freq,
                    self.fft_max_freq,
                    self.sampling_frequency,
                    "first_axis",
                )
            ]
        )
        X_transformed = []
        y_transformed = np.empty((0))
        self.end_index_list = (
            []
        )  # This is the list of end indices for samples generated per seizure

        # The transformation map is just a list of indices corresponding to the last sample generated by each time-series.
        for i in range(len(X)):
            seizure_data = X[i]
            start, step = 0, int(np.floor(self.window_step * self.sampling_frequency))
            stop = start + int(np.floor(self.window_length * self.sampling_frequency))
            fft_data = []
            while stop < seizure_data.shape[1]:
                signal_window = seizure_data[:, start:stop]
                fft_window = pipeline.apply(signal_window)
                fft_data.append(fft_window)
                start, stop = start + step, stop + step
            X_transformed.extend(fft_data)
            if y is not None:
                seizure_label = y[i]
                labels_for_all_seizure_samples = np.full(len(fft_data), seizure_label)
                y_transformed = np.hstack(
                    (y_transformed, labels_for_all_seizure_samples)
                )
            previous_element = self.end_index_list[i - 1] if (i - 1) >= 0 else 0
            self.end_index_list.append(previous_element + len(fft_data))

        X_transformed = np.array(X_transformed)
        if y is None:
            y_transformed = None

        return X_transformed, y_transformed

    def get_transform_meta_output(self):
        if self.end_index_list is not None:
            return {"end_index_list": self.end_index_list}
        else:
            raise ValueError(
                "Must call transform before trying to access its meta output."
            )


_hyperparams_schema = {
    "description": "TODO",
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": [
                "window_length",
                "window_step",
                "fft_min_freq",
                "fft_max_freq",
                "sampling_frequency",
            ],
            "relevantToOptimizer": ["window_length", "window_step", "fft_max_freq"],
            "properties": {
                "window_length": {
                    "type": "number",
                    "default": 1,
                    "description": "TODO",
                    "minimumForOptimizer": 0.25,
                    "maximumForOptimizer": 2,
                    "distribution": "uniform",
                },
                "window_step": {
                    "type": "number",
                    "default": 0.5,
                    "description": "TODO",
                    "minimumForOptimizer": 0.25,
                    "maximumForOptimizer": 1,  # TODO: This is $data->window_length once $data is implemented
                    "distribution": "uniform",
                },
                "fft_min_freq": {
                    "type": "integer",
                    "default": 1,
                    "description": "TODO",
                },
                "fft_max_freq": {
                    "type": "integer",
                    "default": 24,
                    "description": "TODO",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 30,  # TODO: This is $data->sampling_frequency/2 once $data is implemented
                    "distribution": "uniform",
                },
                "sampling_frequency": {
                    "type": "integer",
                    "default": 250,
                    "description": "TODO",
                },
            },
        }
        # TODO: Any constraints on hyper-parameter combinations?
    ],
}

_input_transform_schema = {
    "description": "Input format for data passed to the transform method.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "array", "items": {"type": "number"}},
            },
            "description": "The input data to complete.",
        },
        "y": {
            "type": "array",
            "items": {"anyOf": [{"type": "integer"}, {"type": "string"}]},
        },
    },
}
_output_transform_schema = {
    "description": "The input data to complete.",
    "type": "array",  # This is actually a tuple of X and y
    "items": {"type": "array"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


TimeFreqEigenVectors = lale.operators.make_operator(
    _TimeFreqEigenVectorsImpl, _combined_schemas
)

lale.docstrings.set_docstrings(TimeFreqEigenVectors)



# Copyright 2019,2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    import snapml  # type: ignore

    snapml_installed = True
except ImportError:
    snapml_installed = False

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators


class _SnapBoostingMachineRegressorImpl:
    def __init__(
        self,
        num_round=100,
        objective="mse",
        learning_rate=0.1,
        random_state=0,
        colsample_bytree=1.0,
        subsample=1.0,
        verbose=False,
        lambda_l2=0.0,
        early_stopping_rounds=10,
        compress_trees=False,
        base_score=None,
        max_depth=None,
        min_max_depth=1,
        max_max_depth=5,
        n_jobs=1,
        use_histograms=True,
        hist_nbins=256,
        use_gpu=False,
        gpu_id=0,
        tree_select_probability=1.0,
        regularizer=1.0,
        fit_intercept=False,
        gamma=1.0,
        n_components=10,
    ):
        assert (
            snapml_installed
        ), """Your Python environment does not have snapml installed. Install using: pip install snapml"""
        self._hyperparams = {
            "num_round": num_round,
            "objective": objective,
            "learning_rate": learning_rate,
            "random_state": random_state,
            "colsample_bytree": colsample_bytree,
            "subsample": subsample,
            "verbose": verbose,
            "lambda_l2": lambda_l2,
            "early_stopping_rounds": early_stopping_rounds,
            "compress_trees": compress_trees,
            "base_score": base_score,
            "max_depth": max_depth,
            "min_max_depth": min_max_depth,
            "max_max_depth": max_max_depth,
            "n_jobs": n_jobs,
            "use_histograms": use_histograms,
            "hist_nbins": hist_nbins,
            "use_gpu": use_gpu,
            "gpu_id": gpu_id,
            "tree_select_probability": tree_select_probability,
            "regularizer": regularizer,
            "fit_intercept": fit_intercept,
            "gamma": gamma,
            "n_components": n_components,
        }
        self._wrapped_model = snapml.SnapBoostingMachineRegressor(**self._hyperparams)

    def fit(self, X, y, **fit_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        y = lale.datasets.data_schemas.strip_schema(y)
        self._wrapped_model.fit(X, y, **fit_params)
        return self

    def predict(self, X, **predict_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        return self._wrapped_model.predict(X, **predict_params)


_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "relevantToOptimizer": [
                "num_round",
                "learning_rate",
                "min_max_depth",
                "max_max_depth",
            ],
            "additionalProperties": False,
            "properties": {
                "num_round": {
                    "type": "integer",
                    "minimum": 1,
                    "minimumForOptimizer": 100,
                    "maximumForOptimizer": 1000,
                    "default": 100,
                    "description": "Number of boosting iterations.",
                },
                "objective": {
                    "enum": ["mse", "cross_entropy"],
                    "default": "mse",
                    "description": "Training objective.",
                },
                "learning_rate": {
                    "type": "number",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 0.3,
                    "distribution": "uniform",
                    "default": 0.1,
                    "description": "Learning rate / shrinkage factor.",
                },
                "random_state": {
                    "type": "integer",
                    "default": 0,
                    "description": "Random seed.",
                },
                "colsample_bytree": {
                    "type": "number",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "maximum": 1.0,
                    "default": 1.0,
                    "description": "Fraction of feature columns used at each boosting iteration.",
                },
                "subsample": {
                    "type": "number",
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "maximum": 1.0,
                    "default": 1.0,
                    "description": "Fraction of training examples used at each boosting iteration.",
                },
                "verbose": {
                    "type": "boolean",
                    "default": False,
                    "description": "Print off information during training.",
                },
                "lambda_l2": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 0.0,
                    "description": "L2-reguralization penalty used during tree-building.",
                },
                "early_stopping_rounds": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 10,
                    "description": "When a validation set is provided, training will stop if the validation loss does not increase after a fixed number of rounds.",
                },
                "compress_trees": {
                    "type": "boolean",
                    "default": False,
                    "description": "Compress trees after training for fast inference.",
                },
                "base_score": {
                    "anyOf": [
                        {
                            "type": "number",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Base score to initialize boosting algorithm. If None then the algorithm will initialize the base score to be the the logit of the probability of the positive class.",
                },
                "max_depth": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "If set, will set min_max_depth = max_depth = max_max_depth",
                },
                "min_max_depth": {
                    "type": "integer",
                    "minimum": 1,
                    "minimumForOptimizer": 1,
                    "maximumForOptimizer": 5,
                    "default": 1,
                    "description": "Minimum max_depth of trees in the ensemble.",
                },
                "max_max_depth": {
                    "type": "integer",
                    "minimum": 1,
                    "minimumForOptimizer": 5,
                    "maximumForOptimizer": 10,
                    "default": 5,
                    "description": "Maximum max_depth of trees in the ensemble.",
                },
                "n_jobs": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 1,
                    "description": "Number of threads to use during training.",
                },
                "use_histograms": {
                    "type": "boolean",
                    "default": True,
                    "description": "Use histograms to accelerate tree-building.",
                },
                "hist_nbins": {
                    "type": "integer",
                    "default": 256,
                    "description": "Number of histogram bins.",
                },
                "use_gpu": {
                    "type": "boolean",
                    "default": False,
                    "description": "Use GPU for tree-building.",
                },
                "gpu_id": {
                    "type": "integer",
                    "default": 0,
                    "description": "Device ID for GPU to use during training.",
                },
                "tree_select_probability": {
                    "type": "number",
                    "minimum": 0.0,
                    "maximum": 1.0,
                    "default": 1.0,
                    "description": "Probability of selecting a tree (rather than a kernel ridge regressor) at each boosting iteration.",
                },
                "regularizer": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 1.0,
                    "description": "L2-regularization penality for the kernel ridge regressor.",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": False,
                    "description": "Include intercept term in the kernel ridge regressor.",
                },
                "gamma": {
                    "type": "number",
                    "minimum": 0.0,
                    "default": 1.0,
                    "description": "Guassian kernel parameter.",
                },
                "n_components": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 10,
                    "description": "Number of components in the random projection.",
                },
            },
        },
        {
            "description": "GPU only supported for histogram-based splits.",
            "anyOf": [
                {"type": "object", "properties": {"use_gpu": {"enum": [False]}}},
                {"type": "object", "properties": {"use_histograms": {"enum": [True]}}},
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Build a boosted ensemble from the training set (X, y).",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "y": {
            "description": "The regression target.",
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
            ],
        },
        "sample_weight": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"enum": [None], "description": "Samples are equally weighted."},
            ],
            "description": "Sample weights.",
            "default": None,
        },
        "X_val": {
            "anyOf": [
                {
                    "type": "array",
                    "description": "The outer array is over validation samples aka rows.",
                    "items": {
                        "type": "array",
                        "description": "The inner array is over features aka columns.",
                        "items": {"type": "number"},
                    },
                },
                {"enum": [None], "description": "No validation set provided."},
            ],
            "default": None,
        },
        "y_val": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"enum": [None], "description": "No validation set provided."},
            ],
            "description": "The validation regression target.",
            "default": None,
        },
        "sample_weight_val": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {
                    "enum": [None],
                    "description": "Validation samples are equally weighted.",
                },
            ],
            "description": "Validation sample weights.",
            "default": None,
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "n_jobs": {
            "type": "integer",
            "minimum": 1,
            "default": 1,
            "description": "Number of threads used to run inference.",
        },
    },
}

_output_predict_schema = {
    "description": "The predicted values.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Boosting machine Regressor`_ from `Snap ML`_.

.. _`Boosting machine Regressor`: https://snapml.readthedocs.io/en/latest/#snapml.BoostingMachineRegressor
.. _`Snap ML`: https://www.zurich.ibm.com/snapml/
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.snapml.snap_boosting_machine_regressor.html",
    "import_from": "snapml",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}


SnapBoostingMachineRegressor = lale.operators.make_operator(
    _SnapBoostingMachineRegressorImpl, _combined_schemas
)

lale.docstrings.set_docstrings(SnapBoostingMachineRegressor)



class _FitSpecProxy:
    def __init__(self, base):
        self._base = base

    def __getattr__(self, item):
        return getattr(self._base, item)

    def get_params(self, deep=True):
        ret = {}
        ret["base"] = self._base
        return ret

    def fit(self, X, y, sample_weight=None, **fit_params):
        return self._base.fit(X, y, sample_weight=sample_weight, **fit_params)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Dict, Optional

import lale.docstrings
import lale.operators
import lale.pretty_print
from lale.lib.lale.optimize_suffix import OptimizeSuffix

from ._common_schemas import schema_estimator

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class _OptimizeLast:

    _suffix_optimizer: lale.operators.Operator

    def __init__(
        self,
        estimator: Optional[lale.operators.TrainedOperator] = None,
        last_optimizer: Optional[lale.operators.Operator] = None,
        optimizer_args=None,
        **kwargs
    ):
        if estimator is None:
            last_estimator = None
            lale_prefix = None
        elif isinstance(estimator, lale.operators.TrainedIndividualOp):
            lale_prefix = None
            last_estimator = estimator.clone()
        else:
            assert isinstance(estimator, lale.operators.TrainedPipeline)
            steps = estimator.steps_list()
            num_steps = len(steps)
            if num_steps == 0:
                last_estimator = None
            else:
                last_estimator = estimator.steps_list()[-1].clone()
            lale_prefix = estimator.remove_last()

        self._suffix_optimizer = OptimizeSuffix(
            prefix=lale_prefix,
            suffix=last_estimator,
            optimizer=last_optimizer,
            optimizer_args=optimizer_args,
            **kwargs
        )

    def __getattr__(self, item):
        return getattr(self._suffix_optimizer.shallow_impl, item)

    def fit(self, X_train, y_train=None, **kwargs):
        return self._suffix_optimizer.fit(X_train, y_train, **kwargs)

    def predict(self, X_eval, **predict_params):
        return self._suffix_optimizer.predict(X_eval, **predict_params)


_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "required": [
                "estimator",
                "last_optimizer",
            ],
            "relevantToOptimizer": [],
            "additionalProperties": True,
            "properties": {
                "estimator": schema_estimator,
                "last_optimizer": {
                    "description": "Lale optimizer.\nIf (default) None is specified, Hyperopt is used.",
                    "anyOf": [
                        {"laleType": "operator", "not": {"enum": [None]}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "optimizer_args": {
                    "description": "Parameters to be passed to the optimizer",
                    "anyOf": [
                        {"type": "object"},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "properties": {"X": {}, "y": {}},
}

_input_predict_schema = {"type": "object", "required": ["X"], "properties": {"X": {}}}

_output_predict_schema: Dict[str, Any] = {}

_combined_schemas = {
    "description": """OptimizeLast is a wrapper around other optimizers, which runs the given optimizer
against the suffix, after transforming the data according to the prefix, and then stitches the result together into
a single trained pipeline.

Examples
--------
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.optimize_last.html",
    "import_from": "lale.lib.lale",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}


OptimizeLast = lale.operators.make_operator(_OptimizeLast, _combined_schemas)

lale.docstrings.set_docstrings(OptimizeLast)



from numpy import inf, nan
from sklearn.compose import TransformedTargetRegressor as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _TransformedTargetRegressorImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for TransformedTargetRegressor    Meta-estimator to regress on a transformed target.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "regressor",
                "transformer",
                "func",
                "inverse_func",
                "check_inverse",
            ],
            "relevantToOptimizer": [],
            "additionalProperties": False,
            "properties": {
                "regressor": {
                    "XXX TODO XXX": "object, default=LinearRegression()",
                    "description": "Regressor object such as derived from ``RegressorMixin``",
                    "enum": [None],
                    "default": None,
                },
                "transformer": {
                    "anyOf": [{"type": "object"}, {"enum": [None]}],
                    "default": None,
                    "description": "Estimator object such as derived from ``TransformerMixin``",
                },
                "func": {
                    "XXX TODO XXX": "function, optional",
                    "description": "Function to apply to ``y`` before passing to ``fit``",
                    "enum": [None],
                    "default": None,
                },
                "inverse_func": {
                    "XXX TODO XXX": "function, optional",
                    "description": "Function to apply to the prediction of the regressor",
                    "enum": [None],
                    "default": None,
                },
                "check_inverse": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to check that ``transform`` followed by ``inverse_transform`` or ``func`` followed by ``inverse_func`` leads to the original targets.",
                },
            },
        },
        {
            "description": "transformer' and functions 'func'/'inverse_func' cannot both be set.",
            "anyOf": [
                {"type": "object", "properties": {"transformer": {"enum": [None]}}},
                {
                    "allOf": [
                        {"type": "object", "properties": {"func": {"enum": [None]}}},
                        {
                            "type": "object",
                            "properties": {"inverse_func": {"enum": [None]}},
                        },
                    ]
                },
            ],
        },
        {
            "description": "When 'func' is provided, 'inverse_func' must also be provided",
            "anyOf": [
                {
                    "allOf": [
                        {
                            "type": "object",
                            "properties": {"transformer": {"not": {"enum": [None]}}},
                        },
                        {
                            "anyOf": [
                                {
                                    "type": "object",
                                    "properties": {"func": {"not": {"enum": [None]}}},
                                },
                                {
                                    "type": "object",
                                    "properties": {
                                        "inverse_func": {"not": {"enum": [None]}}
                                    },
                                },
                            ]
                        },
                    ]
                },
                {
                    "type": "object",
                    "properties": {"transformer": {"not": {"enum": [None]}}},
                },
                {"type": "object", "properties": {"func": {"enum": [None]}}},
                {
                    "type": "object",
                    "properties": {"inverse_func": {"not": {"enum": [None]}}},
                },
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the model according to the given training data.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training vector, where n_samples is the number of samples and n_features is the number of features.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values.",
        },
        "sample_weight": {
            "laleType": "Any",
            "XXX TODO XXX": "array-like, shape (n_samples,) optional",
            "description": "Array of weights that are assigned to individual samples",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the base regressor, applying inverse.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predicted values.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.compose.TransformedTargetRegressor#sklearn-compose-transformedtargetregressor",
    "import_from": "sklearn.compose",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
TransformedTargetRegressor = make_operator(
    _TransformedTargetRegressorImpl, _combined_schemas
)

set_docstrings(TransformedTargetRegressor)



from numpy import inf, nan
from sklearn.linear_model import LarsCV as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _LarsCVImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for LarsCV    Cross-validated Least Angle Regression model.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "fit_intercept",
                "verbose",
                "max_iter",
                "normalize",
                "precompute",
                "cv",
                "max_n_alphas",
                "n_jobs",
                "eps",
                "copy_X",
            ],
            "relevantToOptimizer": [
                "fit_intercept",
                "max_iter",
                "normalize",
                "precompute",
                "cv",
                "max_n_alphas",
                "eps",
                "copy_X",
            ],
            "additionalProperties": False,
            "properties": {
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "whether to calculate the intercept for this model",
                },
                "verbose": {
                    "anyOf": [{"type": "boolean"}, {"type": "integer"}],
                    "default": False,
                    "description": "Sets the verbosity amount",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 500,
                    "description": "Maximum number of iterations to perform.",
                },
                "normalize": {
                    "type": "boolean",
                    "default": True,
                    "description": "This parameter is ignored when ``fit_intercept`` is set to False",
                },
                "precompute": {
                    "anyOf": [
                        {
                            "type": "array",
                            "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                            "XXX TODO XXX": "True | False | 'auto' | array-like",
                            "forOptimizer": False,
                        },
                        {"enum": ["auto"]},
                    ],
                    "default": "auto",
                    "description": "Whether to use a precomputed Gram matrix to speed up calculations",
                },
                "cv": {
                    "description": """Cross-validation as integer or as object that has a split function.
                        The fit method performs cross validation on the input dataset for per
                        trial, and uses the mean cross validation performance for optimization.
                        This behavior is also impacted by handle_cv_failure flag.
                        If integer: number of folds in sklearn.model_selection.StratifiedKFold.
                        If object with split function: generator yielding (train, test) splits
                        as arrays of indices. Can use any of the iterators from
                        https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators.""",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "default": 5,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 4,
                            "distribution": "uniform",
                        },
                        {"laleType": "Any", "forOptimizer": False},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "max_n_alphas": {
                    "type": "integer",
                    "minimumForOptimizer": 1000,
                    "maximumForOptimizer": 1001,
                    "distribution": "uniform",
                    "default": 1000,
                    "description": "The maximum number of points on the path used to compute the residuals in the cross-validation",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "Number of CPUs to use during the cross validation",
                },
                "eps": {
                    "type": "number",
                    "minimumForOptimizer": 0.001,
                    "maximumForOptimizer": 0.1,
                    "distribution": "loguniform",
                    "default": 2.220446049250313e-16,
                    "description": "The machine-precision regularization in the computation of the Cholesky diagonal factors",
                },
                "copy_X": {
                    "type": "boolean",
                    "default": True,
                    "description": "If ``True``, X will be copied; else, it may be overwritten.",
                },
            },
        },
        {"XXX TODO XXX": "Parameter: precompute > only subsets of x"},
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the model using X, y as training data.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values.",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the linear model",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Returns predicted values.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.LarsCV#sklearn-linear_model-larscv",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
LarsCV = make_operator(_LarsCVImpl, _combined_schemas)

set_docstrings(LarsCV)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re
from typing import Dict, Iterable, List, Optional

from hyperopt import hp
from hyperopt.pyll import scope

from lale import helpers
from lale.helpers import make_indexed_name
from lale.operators import Operator
from lale.search.PGO import FrequencyDistribution
from lale.search.search_space import (
    SearchSpace,
    SearchSpaceArray,
    SearchSpaceDict,
    SearchSpaceEmpty,
    SearchSpaceEnum,
    SearchSpaceError,
    SearchSpaceNumber,
    SearchSpaceObject,
    SearchSpaceOperator,
    SearchSpaceProduct,
    SearchSpaceSum,
)
from lale.util.Visitor import Visitor, accept


def search_space_to_hp_expr(space: SearchSpace, name: str):
    return SearchSpaceHPExprVisitor.run(space, name)


def search_space_to_hp_str(space: SearchSpace, name: str) -> str:
    return SearchSpaceHPStrVisitor.run(space, name)


def search_space_to_str_for_comparison(space: SearchSpace, name: str) -> str:
    return SearchSpaceHPStrVisitor.run(space, name, counter=None, useCounter=False)


def _mk_label(label, counter, useCounter=True):
    if counter is None or not useCounter:
        return label
    else:
        return f"{label}{counter}"


@scope.define
def pgo_sample(pgo, sample):
    return pgo[sample]


@scope.define
def make_nested_hyperopt(space):
    return helpers.make_nested_hyperopt_space(space)


class SearchSpaceHPExprVisitor(Visitor):
    names: Dict[str, int]

    @classmethod
    def run(cls, space: SearchSpace, name: str):
        visitor = cls(name)
        return accept(space, visitor, name)

    def __init__(self, name: str):
        super(SearchSpaceHPExprVisitor, self).__init__()
        self.names = {}

    def get_unique_name(self, name: str) -> str:
        if name in self.names:
            counter = self.names[name] + 1
            self.names[name] = counter
            return f"{name}${str(counter)}"
        else:
            self.names[name] = 0
            return name

    def mk_label(self, label, counter, useCounter=True):
        return self.get_unique_name(_mk_label(label, counter, useCounter=useCounter))

    def visitSearchSpaceEnum(self, space: SearchSpaceEnum, path: str, counter=None):
        def as_hp_vals(v):
            # Lists are not "safe" to pass to hyperopt without wrapping
            if isinstance(v, (list, tuple, Operator)):
                return helpers.val_wrapper(v)
            else:
                return v

        if len(space.vals) == 1:
            return as_hp_vals(space.vals[0])
        else:
            return hp.choice(
                self.mk_label(path, counter), [as_hp_vals(v) for v in space.vals]
            )

    visitSearchSpaceConstant = visitSearchSpaceEnum
    visitSearchSpaceBool = visitSearchSpaceEnum

    def visitSearchSpaceNumber(self, space: SearchSpaceNumber, path: str, counter=None):
        label = self.mk_label(path, counter)

        if space.pgo is not None:
            return scope.pgo_sample(
                space.pgo, hp.quniform(label, 0, len(space.pgo) - 1, 1)
            )

        dist = "uniform"
        if space.distribution:
            dist = space.distribution

        if space.maximum is None:
            raise SearchSpaceError(
                path, f"maximum not specified for a number with distribution {dist}"
            )
        max = space.getInclusiveMax()
        # if the maximum is not None, the inclusive maximum should not be none
        assert max is not None

        # These distributions need only a maximum
        if dist == "integer":
            if not space.discrete:
                raise SearchSpaceError(
                    path,
                    "integer distribution specified for a non discrete numeric type",
                )
            return hp.randint(label, max)

        if space.minimum is None:
            raise SearchSpaceError(
                path, f"minimum not specified for a number with distribution {dist}"
            )
        min = space.getInclusiveMin()
        # if the minimum is not None, the inclusive minimum should not be none
        assert min is not None

        if dist == "uniform":
            if space.discrete:
                return scope.int(hp.quniform(label, min, max, 1))
            else:
                return hp.uniform(label, min, max)
        elif dist == "loguniform":
            # for log distributions, hyperopt requires that we provide the log of the min/max
            if min <= 0:
                raise SearchSpaceError(
                    path,
                    f"minimum of 0 specified with a {dist} distribution.  This is not allowed; please set it (possibly using minimumForOptimizer) to be positive",
                )
            if min > 0:
                min = math.log(min)
            if max > 0:
                max = math.log(max)
            if space.discrete:
                return scope.int(hp.qloguniform(label, min, max, 1))
            else:
                return hp.loguniform(label, min, max)

        else:
            raise SearchSpaceError(path, f"Unknown distribution type: {dist}")

    def array_single_expr_(self, space: SearchSpaceArray, path: str, num):
        p = _mk_label(path, num) + "_"
        items: Iterable[SearchSpace] = space.items()
        ret = [accept(sub, self, p, counter=x) for x, sub in enumerate(items)]
        return tuple(ret) if space.is_tuple else ret

    def visitSearchSpaceArray(self, space: SearchSpaceArray, path: str, counter=None):
        assert space.maximum >= space.minimum
        p = _mk_label(path, counter)
        cp = p + "_"

        if space.minimum == space.maximum:
            expr = self.array_single_expr_(space, cp, space.minimum)
            return expr
        else:
            exprs = [
                self.array_single_expr_(space, cp, x)
                for x in range(space.minimum, space.maximum + 1)
            ]
            res = hp.choice(p, exprs)
            return res

    def visitSearchSpaceObject(self, space: SearchSpaceObject, path: str, counter=None):
        search_space = {}
        any_path = self.get_unique_name(_mk_label(path, counter) + "_" + "combos")
        search_space["name"] = space.longName

        child_counter = None

        def asexpr(key, e):
            nonlocal child_counter
            if e is None:
                return None
            else:
                ee = accept(e, self, path + "_" + key, counter=child_counter)
                if child_counter is None:
                    child_counter = 1
                else:
                    child_counter = child_counter + 1
                return ee

        def choice_as_tuple_expr(c):
            assert len(space.keys) == len(c)
            ret = [asexpr(space.keys[ind], e) for ind, e in enumerate(c)]
            return ret

        choices = [choice_as_tuple_expr(c) for c in space.choices]

        valid_hyperparam_combinations = hp.choice(any_path, choices)
        i = 0
        for k in space.keys:
            search_space[k] = valid_hyperparam_combinations[i]
            i = i + 1
        return search_space

    def visitSearchSpaceDict(self, sd: SearchSpaceDict, path: str, counter=None):
        search_spaces = {
            name: accept(space, self, path + "_" + name)
            for name, space in sd.space_dict.items()
        }
        return search_spaces

    def visitSearchSpaceProduct(
        self, prod: SearchSpaceProduct, path: str, counter=None
    ):
        search_spaces = [
            accept(space, self, self.get_unique_name(make_indexed_name(name, index)))
            for name, index, space in prod.get_indexed_spaces()
        ]
        return search_spaces

    def visitSearchSpaceSum(self, sum: SearchSpaceSum, path: str, counter=None):
        if len(sum.sub_spaces) == 1:
            return accept(sum.sub_spaces[0], self, "")
        else:
            unique_name: str = self.get_unique_name("choice")
            search_spaces = hp.choice(
                unique_name,
                [{str(i): accept(m, self, "")} for i, m in enumerate(sum.sub_spaces)],
            )
            return search_spaces

    def visitSearchSpaceOperator(
        self, op: SearchSpaceOperator, path: str, counter=None
    ):
        return scope.make_nested_hyperopt(accept(op.sub_space, self, path))

    def visitSearchSpaceEmpty(self, op: SearchSpaceEmpty, path: str, counter=None):
        raise SearchSpaceError(
            path, "The hyperopt backend can't compile an empty (sub-) search space"
        )


class SearchSpaceHPStrVisitor(Visitor):
    pgo_dict: Dict[str, FrequencyDistribution]
    names: Dict[str, int]

    pgo_header: Optional[str]
    nested_header: Optional[str]
    decls: str

    @classmethod
    def run(cls, space: SearchSpace, name: str, counter=None, useCounter=True):
        visitor = cls(name)
        ret: str = ""
        body = accept(space, visitor, name, counter=counter, useCounter=useCounter)
        if visitor.pgo_header is not None:
            ret += visitor.pgo_header
        if visitor.nested_header is not None:
            ret += visitor.nested_header
        if visitor.decls:
            ret += visitor.decls + "\n"
        ret += "return " + body
        return ret

    def get_unique_name(self, name: str) -> str:
        if name in self.names:
            counter = self.names[name] + 1
            self.names[name] = counter
            return f"{name}${str(counter)}"
        else:
            self.names[name] = 0
            return name

    def get_unique_variable_name(self, name: str) -> str:
        if name in self.names:
            counter = self.names[name] + 1
            self.names[name] = counter
            return f"{name}__{str(counter)}"
        else:
            self.names[name] = 0
            return name

    def mk_label(self, label, counter, useCounter=True):
        return self.get_unique_name(_mk_label(label, counter, useCounter=useCounter))

    def __init__(self, name: str):
        super(SearchSpaceHPStrVisitor, self).__init__()
        self.pgo_dict = {}
        self.names = {}
        self.pgo_header = None
        self.nested_header = None
        self.decls = ""

    def visitSearchSpaceEnum(
        self, space: SearchSpaceEnum, path: str, counter=None, useCounter=True
    ):
        def val_as_str(v):
            if v is None:
                return "null"
            elif isinstance(v, str):
                return f"'{v}'"
            else:
                return str(v)

        if len(space.vals) == 1:
            return val_as_str(space.vals[0])
        else:
            vals_str = "[" + ", ".join([val_as_str(v) for v in space.vals]) + "]"
            return (
                f"hp.choice('{self.mk_label(path, counter, useCounter)}', {vals_str})"
            )

    visitSearchSpaceConstant = visitSearchSpaceEnum
    visitSearchSpaceBool = visitSearchSpaceEnum

    def visitSearchSpaceNumber(
        self, space: SearchSpaceNumber, path: str, counter=None, useCounter=True
    ):
        label = self.mk_label(path, counter, useCounter=useCounter)

        if space.pgo is not None:
            self.pgo_dict[label] = space.pgo
            return f"scope.pgo_sample(pgo_{label}, hp.quniform('{label}', {0}, {len(space.pgo)-1}, 1))"

        dist = "uniform"
        if space.distribution:
            dist = space.distribution

        if space.maximum is None:
            raise SearchSpaceError(
                path, f"maximum not specified for a number with distribution {dist}"
            )

        max = space.getInclusiveMax()
        # if the maximum is not None, the inclusive maximum should not be none
        assert max is not None

        # These distributions need only a maximum
        if dist == "integer":
            if not space.discrete:
                raise SearchSpaceError(
                    path,
                    "integer distribution specified for a non discrete numeric type....",
                )

            return f"hp.randint('{label}', {max})"

        if space.minimum is None:
            raise SearchSpaceError(
                path, f"minimum not specified for a number with distribution {dist}"
            )
        min = space.getInclusiveMin()
        # if the minimum is not None, the inclusive minimum should not be none
        assert min is not None

        if dist == "uniform":
            if space.discrete:
                return f"hp.quniform('{label}', {min}, {max}, 1)"
            else:
                return f"hp.uniform('{label}', {min}, {max})"
        elif dist == "loguniform":
            # for log distributions, hyperopt requires that we provide the log of the min/max
            if min <= 0:
                raise SearchSpaceError(
                    path,
                    f"minimum of 0 specified with a {dist} distribution.  This is not allowed; please set it (possibly using minimumForOptimizer) to be positive",
                )
            if min > 0:
                min = math.log(min)
            if max > 0:
                max = math.log(max)

            if space.discrete:
                return f"hp.qloguniform('{label}', {min}, {max}, 1)"
            else:
                return f"hp.loguniform('{label}', {min}, {max})"
        else:
            raise SearchSpaceError(path, f"Unknown distribution type: {dist}")

    def array_single_str_(
        self, space: SearchSpaceArray, path: str, num, useCounter=True
    ) -> str:
        p = _mk_label(path, num, useCounter=useCounter) + "_"
        ret = "(" if space.is_tuple else "["
        items: Iterable[SearchSpace] = space.items()
        ret += ",".join(
            (
                accept(sub, self, p, counter=x, useCounter=useCounter)
                for x, sub in enumerate(items)
            )
        )
        ret += ")" if space.is_tuple else "]"
        return ret

    def visitSearchSpaceArray(
        self, space: SearchSpaceArray, path: str, counter=None, useCounter=True
    ) -> str:
        assert space.maximum >= space.minimum
        p = _mk_label(path, counter, useCounter=useCounter)
        cp = p + "_"

        if space.minimum == space.maximum:
            return self.array_single_str_(
                space, cp, space.minimum, useCounter=useCounter
            )
        else:
            res = "hp.choice(" + p + ", ["
            res += ",".join(
                (
                    self.array_single_str_(space, cp, x, useCounter=useCounter)
                    for x in range(space.minimum, space.maximum + 1)
                )
            )
            res += "])"
            return res

    def visitSearchSpaceObject(
        self, space: SearchSpaceObject, path: str, counter=None, useCounter=True
    ):
        s_decls = []
        space_name = self.get_unique_variable_name("search_space")
        any_name = self.get_unique_variable_name("valid_hyperparam_combinations")
        any_path = self.get_unique_name(
            _mk_label(path, counter, useCounter=useCounter) + "_" + "combos"
        )
        s_decls.append(space_name + " = {}")
        s_decls.append(f"{space_name}['name'] = {space.longName}")
        child_counter = None

        def cstr(key, x):
            nonlocal child_counter
            if x is None:
                return "None"
            else:
                s = accept(
                    x, self, path + "_" + key, child_counter, useCounter=useCounter
                )
                if child_counter is None:
                    child_counter = 1
                else:
                    child_counter = child_counter + 1
                return s

        def choice_as_tuple_str(c):
            assert len(space.keys) == len(c)
            ret = [cstr(space.keys[ind], e) for ind, e in enumerate(c)]
            return ret

        str_choices = (
            "["
            + ",".join(
                ["(" + ",".join(choice_as_tuple_str(c)) + ")" for c in space.choices]
            )
            + "]"
        )
        s_decls.append(f"{any_name} = hp.choice('{any_path}', {str_choices})")
        i = 0
        for k in space.keys:
            s_decls.append(f"{space_name}['{k}'] = {any_name}[{i}]")
            i = i + 1

        if self.pgo_dict:
            if not self.pgo_header:
                self.pgo_header = """
@scope.define
def pgo_sample(pgo, sample):
    return pgo[sample]

"""

            # use this to sort the pgo_labels by the unique ind
            # appended to the key.
            # This puts them in the order they appear in the hyperopt
            # expression, making it easier to read
            def last_num(kv):
                matches = re.search(r"(\d+)$", kv[0])
                if matches is None:
                    return 0
                else:
                    return int(matches.group(0))

            pgo_decls: List[str] = []
            for k, v in sorted(self.pgo_dict.items(), key=last_num):
                fl = v.freq_dist.tolist()
                pgo_decls.append(f"pgo_{k} = {fl}")
            if self.decls:
                self.decls = self.decls + "\n"
            self.decls = self.decls + "\n".join(pgo_decls)
        self.decls += "\n".join(s_decls) + "\n"
        return space_name

    def visitSearchSpaceDict(
        self, sd: SearchSpaceDict, path: str, counter=None, useCounter=True
    ):
        search_spaces = (
            name + ":" + accept(space, self, path + "_" + name)
            for name, space in sd.space_dict.items()
        )
        return "{" + ",".join(search_spaces) + "}"

    def visitSearchSpaceProduct(
        self, prod: SearchSpaceProduct, path: str, counter=None, useCounter=True
    ):
        search_spaces = (
            accept(space, self, self.get_unique_name(make_indexed_name(name, index)))
            for name, index, space in prod.get_indexed_spaces()
        )
        return "[" + ",".join(search_spaces) + "]"

    def visitSearchSpaceSum(
        self, sum_space: SearchSpaceSum, path: str, counter=None, useCounter=True
    ):
        unique_name: str = self.get_unique_name("choice")
        sub_str: Iterable[str] = (
            '"' + str(i) + '"' + " : " + '"' + accept(m, self, "") + '"'
            for i, m in enumerate(sum_space.sub_spaces)
        )
        sub_spaces_str: str = "[" + ",".join(sub_str) + "]"
        return f"hp.choice({unique_name}, {sub_spaces_str})"

    def visitSearchSpaceOperator(
        self, op: SearchSpaceOperator, path: str, counter=None, useCounter=True
    ):
        if not self.nested_header:
            self.nested_header = """
@scope.define
def make_nested_hyperopt(space):
    from lale.helpers import make_nested_hyperopt_space
    return make_nested_hyperopt_space(space)

"""
        return f"scope.make_nested_hyperopt({accept(op.sub_space, self, path, counter=counter, useCounter=useCounter)})"

    def visitSearchSpaceEmpty(
        self, op: SearchSpaceEmpty, path: str, counter=None, useCounter=True
    ) -> str:
        return "***EMPTY**"



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

JSON_TYPE = Dict[str, Any]

schema_estimator: JSON_TYPE = {
    "description": "Planned Lale individual operator or pipeline.",
    "anyOf": [
        {"laleType": "operator"},
        {
            "enum": [None],
            "description": "lale.lib.sklearn.LogisticRegression",
        },
    ],
    "default": None,
}

# schemas used by many optimizers
schema_scoring_item: JSON_TYPE = {
    "description": "Scorer object, or known scorer named by string.",
    "anyOf": [
        {
            "description": """Callable with signature ``scoring(estimator, X, y)`` as documented in `sklearn scoring`_.

The callable has to return a scalar value, such that a higher score is better.
This may be created from one of the `sklearn metrics`_ using `make_scorer`_.
Or it can be one of the scoring callables returned by the factory
functions in `lale.lib.aif360 metrics`_, for example,
``symmetric_disparate_impact(**fairness_info)``.
Or it can be a completely custom user-written Python callable.

.. _`sklearn scoring`: https://scikit-learn.org/stable/modules/model_evaluation.html#the-scoring-parameter-defining-model-evaluation-rules
.. _`make_scorer`: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer.
.. _`sklearn metrics`: https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics
.. _`lale.lib.aif360 metrics`: https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.html#metrics
""",
            "laleType": "callable",
        },
        {
            "description": "Known scorer for classification task.",
            "enum": [
                "accuracy",
                "explained_variance",
                "max_error",
                "roc_auc",
                "roc_auc_ovr",
                "roc_auc_ovo",
                "roc_auc_ovr_weighted",
                "roc_auc_ovo_weighted",
                "balanced_accuracy",
                "average_precision",
                "neg_log_loss",
                "neg_brier_score",
            ],
        },
        {
            "description": "Known scorer for regression task.",
            "enum": [
                "r2",
                "neg_mean_squared_error",
                "neg_mean_absolute_error",
                "neg_root_mean_squared_error",
                "neg_mean_squared_log_error",
                "neg_median_absolute_error",
            ],
        },
    ],
}

schema_scoring_single: JSON_TYPE = {
    "description": "Scorer object, or known scorer named by string.",
    "anyOf": [
        {
            "enum": [None],
            "description": "When not specified, use `accuracy` for classification tasks and `r2` for regression.",
        },
        schema_scoring_item,
    ],
}

schema_scoring_list: JSON_TYPE = {
    "description": "A list of Scorer objects, or known scorers named by string.  The optimizer may take the order into account.",
    "type": "array",
    "items": schema_scoring_item,
}

schema_scoring: JSON_TYPE = {
    "description": "Either a single or a list of (Scorer objects, or known scorers named by string).",
    "anyOf": [schema_scoring_single, schema_scoring_list],
}

schema_best_score_single: JSON_TYPE = {
    "description": """The best score for the specified scorer.

Given that higher scores are better, passing ``(best_score - score)``
as a loss to the minimizing optimizer will maximize the score.
By specifying best_score, the loss can be >=0, where 0 is the best loss.""",
    "type": "number",
    "default": 0.0,
}

schema_best_score: JSON_TYPE = {
    "description": """The best score for the specified scorer.

Given that higher scores are better, passing ``(best_score - score)``
as a loss to the minimizing optimizer will maximize the score.
By specifying best_score, the loss can be >=0, where 0 is the best loss.""",
    "default": 0.0,
    "anyOf": [
        schema_best_score_single,
        {
            "description": """The best score for each specified scorer.
            If not enough are specified, the remainder are assumed to be the default.

            Given that higher scores are better, passing ``(best_score - score)``
            as a loss to the minimizing optimizer will maximize the score.
            By specifying best_score, the loss can be >=0, where 0 is the best loss.""",
            "type": "array",
            "items": schema_best_score_single,
        },
    ],
}


def check_scoring_best_score_constraint(scoring=None, best_score=0) -> None:
    if isinstance(best_score, list):
        if isinstance(scoring, list):
            if len(scoring) < len(best_score):
                raise ValueError(
                    f"Error: {len(best_score)} best scores were specified, but there are only {len(scoring)} scorers."
                )
        else:
            raise ValueError(
                f"Error: {len(best_score)} best scores were specified, but there is only one scorer."
            )


schema_simple_cv: JSON_TYPE = {
    "description": "Number of folds for cross-validation.",
    "type": "integer",
    "minimum": 2,
    "default": 5,
    "minimumForOptimizer": 3,
    "maximumForOptimizer": 4,
    "distribution": "uniform",
}

schema_cv: JSON_TYPE = {
    "description": """Cross-validation as integer or as object that has a split function.

The fit method performs cross validation on the input dataset for per
trial, and uses the mean cross validation performance for optimization.
This behavior is also impacted by the handle_cv_failure flag.
""",
    "anyOf": [
        schema_simple_cv,
        {
            "not": {"type": "integer"},
            "forOptimizer": False,
            "description": "Object with split function: generator yielding (train, test) splits as arrays of indices. Can use any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators",
        },
    ],
    "default": 5,
}

schema_max_opt_time: JSON_TYPE = {
    "description": "Maximum amount of time in seconds for the optimization.",
    "anyOf": [
        {"type": "number", "minimum": 0.0},
        {"description": "No runtime bound.", "enum": [None]},
    ],
    "default": None,
}



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

schema_X_numbers_y_top = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {"description": "Target class labels (unused)."},
    },
}

schema_X_numbers = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
    },
}

schema_1D_cats = {
    "anyOf": [
        {"type": "array", "items": {"type": "string"}},
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "boolean"}},
    ],
}

schema_1D_numbers = {"type": "array", "items": {"type": "number"}}

schema_2D_numbers = {
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}



import os
import unittest
import urllib.request
from typing import Optional

import joblib
import pandas as pd
import sklearn
from sklearn.linear_model import LogisticRegression as LR
from sklearn.neighbors import KNeighborsClassifier as KNN
from sklearn.tree import DecisionTreeClassifier as Tree

import lale
import lale.operators
from lale.helpers import println_pos
from lale.lib.autoai_libs import wrap_pipeline_segments

assert sklearn.__version__ == "0.23.1", "This test is for scikit-learn 0.23.1."


class TestAutoAIOutputConsumption(unittest.TestCase):

    pickled_model_path = os.path.join(
        os.path.dirname(__file__),
        "..",
        "lale",
        "datasets",
        "autoai",
        "credit_risk.pickle",
    )
    train_csv_path = "german_credit_data_biased_training.csv"
    train_csv_url = "https://raw.githubusercontent.com/pmservice/wml-sample-models/master/autoai/credit-risk-prediction/data/german_credit_data_biased_training.csv"
    training_df = None
    model = None
    prefix_model = None
    refined_model = None
    pipeline_content: Optional[str] = None
    pp_pipeline: Optional[lale.operators.TrainablePipeline] = None

    @classmethod
    def setUp(cls) -> None:
        urllib.request.urlretrieve(cls.train_csv_url, cls.train_csv_path)
        TestAutoAIOutputConsumption.training_df = pd.read_csv(
            TestAutoAIOutputConsumption.train_csv_path
        )

    def test_01_load_pickled_model(self):
        try:
            old_model = joblib.load(TestAutoAIOutputConsumption.pickled_model_path)
            TestAutoAIOutputConsumption.model = old_model
            println_pos(f"type(model) {type(TestAutoAIOutputConsumption.model)}")
            println_pos(f"model {str(TestAutoAIOutputConsumption.model)}")
        except Exception as e:
            assert False, f"Exception was thrown during model pickle: {e}"

    def test_02_predict_on_model(self):
        t_df = TestAutoAIOutputConsumption.training_df
        assert t_df is not None
        check_X = t_df.drop(["Risk"], axis=1).values
        x = check_X
        try:
            m = TestAutoAIOutputConsumption.model
            assert m is not None
            pred = m.predict(x)
            assert len(pred) == len(
                x
            ), f"Prediction has returned unexpected number of rows {len(pred)} - expected {len(x)}"
        except Exception as e:
            assert False, f"Exception was thrown during model prediction: {e}"

    def test_03_print_pipeline(self):
        lale_pipeline = TestAutoAIOutputConsumption.model
        assert lale_pipeline is not None
        wrapped_pipeline = wrap_pipeline_segments(lale_pipeline)
        assert wrapped_pipeline is not None
        TestAutoAIOutputConsumption.pipeline_content = wrapped_pipeline.pretty_print()
        assert type(TestAutoAIOutputConsumption.pipeline_content) is str
        assert len(TestAutoAIOutputConsumption.pipeline_content) > 0
        println_pos(
            f'pretty-printed """{TestAutoAIOutputConsumption.pipeline_content}"""'
        )
        assert (
            "lale.wrap_imported_operators()"
            in TestAutoAIOutputConsumption.pipeline_content
        )

    def test_04_execute_pipeline(self):
        try:
            with open("pp_pipeline.py", "w") as pipeline_f:
                assert TestAutoAIOutputConsumption.pipeline_content is not None
                pipeline_f.write(TestAutoAIOutputConsumption.pipeline_content)

            import importlib.util

            import lale.operators

            spec = importlib.util.spec_from_file_location(
                "pp_pipeline", "pp_pipeline.py"
            )
            assert spec is not None
            pipeline_module = importlib.util.module_from_spec(spec)
            assert spec.loader is not None
            # the type stubs for _Loader are currently incomplete
            spec.loader.exec_module(pipeline_module)  # type: ignore
            TestAutoAIOutputConsumption.pp_pipeline = pipeline_module.pipeline  # type: ignore
            assert isinstance(
                TestAutoAIOutputConsumption.pp_pipeline,
                lale.operators.TrainablePipeline,
            )
        except Exception as e:
            assert False, f"{e}"
        finally:
            try:
                os.remove("pp_pipeline.py")
            except OSError:
                println_pos("Couldn't remove pp_pipeline.py file")

    def test_05_train_pretty_print_pipeline(self):
        t_df = TestAutoAIOutputConsumption.training_df
        assert t_df is not None
        train_X = t_df.drop(["Risk"], axis=1).values
        train_y = t_df.Risk.values

        ppp = TestAutoAIOutputConsumption.pp_pipeline
        assert ppp is not None
        ppp.fit(train_X, train_y)

    def test_06_predict_on_pretty_print_pipeline(self):
        t_df = TestAutoAIOutputConsumption.training_df
        assert t_df is not None

        check_X = t_df.drop(["Risk"], axis=1).values
        x = check_X
        try:
            ppp = TestAutoAIOutputConsumption.pp_pipeline
            assert ppp is not None

            pred = ppp.predict(x)
            assert len(pred) == len(
                x
            ), f"Prediction on pretty print model has returned unexpected number of rows {len(pred)} - expected {len(x)}"
        except Exception as e:
            assert (
                False
            ), f"Exception was thrown during pretty print model prediction: {e}"

    def test_07_convert_model_to_lale(self):
        import lale.operators

        try:
            lale_pipeline = TestAutoAIOutputConsumption.model
            assert lale_pipeline is not None
            TestAutoAIOutputConsumption.prefix_model = (
                lale_pipeline.remove_last().freeze_trainable()
            )
            assert isinstance(
                TestAutoAIOutputConsumption.prefix_model,
                lale.operators.TrainablePipeline,
            )
        except Exception as e:
            assert False, f"Exception was thrown during model prediction: {e}"

    def test_08_refine_model_with_lale(self):
        from lale import wrap_imported_operators
        from lale.lib.lale import Hyperopt

        wrap_imported_operators()
        try:
            println_pos(
                f"type(prefix_model) {type(TestAutoAIOutputConsumption.prefix_model)}"
            )
            println_pos(f"type(LR) {type(LR)}")
            # This is for classifiers, regressors needs to have different operators & different scoring metrics (e.g 'r2')
            pm = TestAutoAIOutputConsumption.prefix_model
            assert pm is not None
            new_model = pm >> (LR | Tree | KNN)
            t_df = TestAutoAIOutputConsumption.training_df
            assert t_df is not None
            train_X = t_df.drop(["Risk"], axis=1).values
            train_y = t_df["Risk"].values
            hyperopt = Hyperopt(
                estimator=new_model, cv=2, max_evals=3, scoring="roc_auc"
            )
            hyperopt_pipelines = hyperopt.fit(train_X, train_y)
            TestAutoAIOutputConsumption.refined_model = (
                hyperopt_pipelines.get_pipeline()
            )
        except Exception as e:
            assert False, f"Exception was thrown during model refinery: {e}"

    def test_09_predict_refined_model(self):
        t_df = TestAutoAIOutputConsumption.training_df
        assert t_df is not None
        check_X = t_df.drop(["Risk"], axis=1).values
        x = check_X
        try:
            model = TestAutoAIOutputConsumption.refined_model
            assert model is not None
            pred = model.predict(x)
            assert len(pred) == len(
                x
            ), f"Prediction on refined model has returned unexpected number of rows {len(pred)} - expected {len(x)}"
        except Exception as e:
            assert False, f"Exception was thrown during refined model prediction: {e}"
